import netbook
import asyncio


async def test_text_area(pilot, mocker):
    app: netbook.JupyterTextualApp = pilot.app
    await app.action_run_cell_select_below()
    await pilot.pause()
    assert len(app.cells) == 2
    assert app.focused_cell_id == 1
    assert app.cells[1].has_class("input_focused")

    # Test moving up from text area
    await pilot.press("up")
    await pilot.pause()
    assert app.focused_cell_id == 0

    await app.action_run_cell_and_insert_below()
    await pilot.pause()
    assert len(app.cells) == 3
    assert app.focused_cell_id == 1
    assert app.cells[1].has_class("input_focused")

    # Test moving down from text area
    await pilot.press("down")
    await pilot.pause()
    assert app.focused_cell_id == 2

    source: netbook._text_area.CodeTextArea = app.cells[2].source
    await pilot.press("a")
    assert source.text == "a"

    # Test autocomplete on single suggestion
    app.kernel_client._recv_reply = mocker.AsyncMock(return_value={"content": {"cursor_start": 0, "matches": ["asdf"]}})
    await pilot.press("tab")
    assert not source.completer.visible
    assert source.text == "asdf"

    # Test multiple suggestions
    app.kernel_client._recv_reply = mocker.AsyncMock(
        return_value={"content": {"cursor_start": 0, "matches": ["asdfasdf", "asdffdsa"]}}
    )
    await pilot.press("tab")
    assert source.completer.visible
    assert source.text == "asdf"

    # Test the pressing some keys doesn't close the completer
    await pilot.press("down")
    await pilot.pause()
    assert source.completer.visible

    # Test clicking on an option
    await pilot.click(source.completer, offset=(1, 1))
    await pilot.pause()
    assert not source.completer.visible
    assert source.text == "asdffdsa"

    # Test dismissing completer on unrecognized key
    await pilot.press("tab")
    await pilot.pause()
    assert source.completer.visible
    await pilot.press("left")
    await pilot.pause()
    assert not source.completer.visible

    # Test reopening completer on printable character
    await pilot.press("q")
    await pilot.press("tab")
    await pilot.pause()
    assert app.kernel_client._recv_reply.call_count == 3
    assert source.text == "asdffdsqa"
    assert source.completer.visible

    # Test empty response from kernel
    app.kernel_client._recv_reply = mocker.AsyncMock(return_value={"content": {"cursor_start": 0, "matches": []}})
    await pilot.press("q")
    await pilot.press("tab")
    await pilot.pause()
    assert not source.completer.visible
    assert source.text == "asdffdsqqa"
    assert source.has_focus

    # Test showing inspector
    app.kernel_client._recv_reply = mocker.AsyncMock(
        # long response will force scrollbars
        return_value={"content": {"data": {"text/plain": ("a" * 90 + "\n") * 40}}}
    )
    await pilot.press("shift+tab")
    await pilot.pause()
    assert source.inspect.visible
    await pilot.press("left")
    await pilot.pause()
    assert source.inspect.visible
    await pilot.press("a")
    await pilot.pause()
    assert not source.inspect.visible

    # Test no inspect from kernel
    app.kernel_client._recv_reply = mocker.AsyncMock(return_value={"content": {"data": {}}})
    await pilot.press("shift+tab")
    await pilot.pause()
    assert not source.inspect.visible

    # Test kernel crash / bad response
    app.kernel_client._recv_reply = mocker.AsyncMock(return_value="asdf")
    await pilot.press("shift+tab")
    await pilot.pause()
    assert not source.inspect.visible

    # Test clicking outside
    app.kernel_client._recv_reply = mocker.AsyncMock(return_value={"content": {"data": {"text/plain": "help"}}})
    await pilot.press("shift+tab")
    await pilot.pause()
    assert source.inspect.visible
    assert source.inspect.has_focus
    await pilot.click(source, offset=(0, 0))
    await pilot.pause()
    assert not source.inspect.visible

    # Test kernel busy
    app.kernel_access_queue.add(asyncio.Future())
    app.kernel_client._recv_reply = mocker.AsyncMock()
    await pilot.press("a")  # move the cursor from the start
    await pilot.press("shift+tab")
    await pilot.pause()
    assert not source.inspect.visible
    assert app.kernel_client._recv_reply.call_count == 0
    app.kernel_access_queue = {}



import netbook

import nbformat
import pytest


@pytest.fixture
async def pilot(mocker):
    nb = nbformat.v4.new_notebook(cells=[nbformat.v4.new_code_cell()])
    km = mocker.Mock()
    km.kernel_spec.language = "python"
    kc = mocker.Mock()
    kc.execute_interactive = mocker.AsyncMock()
    app = netbook.JupyterTextualApp(km, kc, "", nb)
    async with app.run_test() as pilot:
        await pilot.pause()
        yield pilot


@pytest.fixture
async def pilot_nb():
    nb_app = netbook.JupyterNetbook()
    nb_app.initialize(["./tests/test.ipynb"])
    assert hasattr(nb_app, "textual_app")
    app = nb_app.textual_app
    async with app.run_test() as pilot:
        await pilot.pause()
        yield pilot



%load_ext rich

# Set the number of columns for tqdm. A bit arbitrary but works better than the default.
from os import environ as _environ
_environ['TQDM_NCOLS'] = '80'

from pandas import DataFrame as _DataFrame
from pandas import set_option as _set_option

_set_option("display.expand_frame_repr", True)

def _pd_rich_repr(self, console, console_options):
    from rich.table import Table
    from rich.box import SIMPLE

    from pandas._config.config import get_option
    from pandas.io.formats.format import DataFrameFormatter
    from pandas.io.formats.string import StringFormatter

    formatter = DataFrameFormatter(
        self,
        max_rows=get_option("display.max_rows"),
        max_cols=get_option("display.max_columns"),
        min_rows=get_option("display.min_rows"),
        show_dimensions=get_option("display.show_dimension"),
    )
    sformatter = StringFormatter(formatter)
    string_cols = sformatter._get_strcols()

    table = Table(box=SIMPLE, row_styles=["dim", ""])
    if formatter.has_index_names:
        for i, c in enumerate(string_cols):
            table.add_column("\n".join(c[:2]), style="bold" if i == 0 else None, justify="left" if i == 0 else "right")
        start_row = 2
    else:
        for i, c in enumerate(string_cols):
            table.add_column(c[0], style="bold" if i == 0 else None, justify="left" if i == 0 else "right")
        start_row = 1
    for i in range(start_row, len(string_cols[0])):
        table.add_row(*[c[i] for c in string_cols])
    if get_option("display.expand_frame_repr"):
        console_options = console_options.update(max_width=256)
    yield from table.__rich_console__(console, console_options)
    if formatter.should_show_dimensions:
        yield formatter.dimensions_info.strip()


_DataFrame.__rich_console__ = _pd_rich_repr



import pathlib
import sys

import jupyter_client.consoleapp
import jupyter_core.application
import jupyter_core.utils

import nbformat

import textual

import traitlets

from netbook._textual_app import JupyterTextualApp


class JupyterNetbook(jupyter_core.application.JupyterApp, jupyter_client.consoleapp.JupyterConsoleApp):
    name = "jupyter-netbook"
    description = """
        A terminal-based app for jupyter notebooks.
    """
    flags = traitlets.Dict({**jupyter_core.application.base_flags, **jupyter_client.consoleapp.app_flags})
    aliases = traitlets.Dict({**jupyter_core.application.base_aliases, **jupyter_client.consoleapp.app_aliases})

    kernel_client_class = jupyter_client.asynchronous.AsyncKernelClient
    # kernel_manager_class = jupyter_client.manager.AsyncKernelManager

    log = textual.log

    def initialize(self, argv):
        if not hasattr(self.log, "exception"):
            # Jupyter requires self.log.exception
            self.log.exception = self.log.error

        if sys.platform == "win32":
            # see https://github.com/zeromq/pyzmq/issues/1423
            import asyncio

            asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

        jupyter_core.application.JupyterApp.initialize(self, argv)
        if self._dispatching:
            return

        nb = None
        nbkernel = None
        # Determine the notebook name and content
        if len(self.extra_args) > 0:
            # TODO: what if multiple filenames are passed
            nbfile = self.extra_args[0]
            if pathlib.Path(nbfile).exists():
                nb = nbformat.read(nbfile, nbformat.current_nbformat)
                if "kernelspec" in nb.metadata:
                    nbkernel = nb.metadata.kernelspec.name
        else:
            # Determine a name like "Untitled{x}.ipynb"
            suffix = ""
            while pathlib.Path(f"Untitled{suffix}.ipynb").exists():
                suffix = suffix + 1 if suffix else 1
            nbfile = f"Untitled{suffix}.ipynb"

        if nbkernel and (
            "JupyterNetbook" not in self.cli_config or "kernel_name" not in self.cli_config["JupyterNetbook"]
        ):
            # If kernel_name is not explicitly specify but is present in the notebook, use that one
            self.kernel_name = nbkernel

        jupyter_client.consoleapp.JupyterConsoleApp.initialize(self, argv)

        # TODO: if --existing is specified, then self.kernel_manager is None.
        self.textual_app = JupyterTextualApp(self.kernel_manager, self.kernel_client, nbfile, nb)

    def start(self):
        jupyter_core.application.JupyterApp.start(self)
        self.textual_app.run()


def main(argv=None, **kwargs) -> None:
    JupyterNetbook.launch_instance(argv, **kwargs)



from __future__ import annotations

import abc
import asyncio
import dataclasses
import re
import typing as tp

import rich.text

import textual.app
import textual.binding
import textual.containers
import textual.reactive
import textual.widgets

if tp.TYPE_CHECKING:
    from netbook import JupyterTextualApp


class PopUpMixin:
    # TODO: unable to declare "visisble" here due to multiple inheritance?
    def watch_visible(self, visible: True) -> None:
        if visible:
            self.remove_class("hidden")
            self.focus(scroll_visible=False)
        else:
            self.add_class("hidden")

    def _on_focus(self, event: textual.events.Focus) -> None:
        self.parent.has_focus = True

    def _on_blur(self, event: textual.events.Blur) -> None:
        if self.visible:
            self.parent.has_focus = False
            self.visible = False

    def action_dismiss(self) -> None:
        self.parent.has_focus = True
        self.visible = False
        self.parent.focus(scroll_visible=False)

    @abc.abstractmethod
    def estimate_height(self) -> int:
        pass


class Completer(textual.widgets.OptionList, PopUpMixin):
    visible: textual.reactive.reactive[bool] = textual.reactive.reactive(False)

    BINDINGS = [
        textual.binding.Binding("backspace, escape", "dismiss", "Dismiss"),
    ]

    def get_handle_keys(self) -> set[str]:
        return {"backspace", "down", "end", "enter", "escape", "home", "pagedown", "pageup", "up"}

    def __init__(self) -> None:
        super().__init__(self, classes="hidden")

    def action_select(self) -> None:
        super().action_select()
        self.visible = False
        self.parent.focus(scroll_visible=False)

    @tp.override
    def estimate_height(self) -> int:
        # TODO: add border shmorder
        assert self.styles.max_height.is_cells
        return min(self.styles.max_height.value, self.option_count)


class Inspect(textual.containers.ScrollableContainer, PopUpMixin, can_focus=True):
    visible: textual.reactive.reactive[bool] = textual.reactive.reactive(False)

    BINDINGS = [
        textual.binding.Binding("escape", "dismiss", "Dismiss"),
    ]

    def __init__(self) -> None:
        super().__init__(classes="hidden")
        self.contents = textual.widgets.Static()

    def compose(self) -> textual.app.ComposeResult:
        yield self.contents

    def update(self, contents: textual.visual.Visual) -> None:
        self.scroll_home(animate=False, immediate=True)
        self.contents.update(contents)

    def get_handle_keys(self) -> set[str]:
        keys = {"escape"}
        if self.allow_vertical_scroll:
            keys.update(["up", "down", "pageup", "pagedown", "home", "end"])
        if self.allow_horizontal_scroll:
            keys.update(["left", "right", "ctrl+pageup", "ctrl+pagedown", "home", "end"])
        return keys

    @tp.override
    def estimate_height(self) -> int:
        assert self.styles.max_height.is_cells
        return min(
            self.styles.max_height.value,
            len(str(self.contents.renderable).splitlines())
            + self.styles.border.spacing.top
            + self.styles.border.spacing.bottom,
        )


class CellTextArea(textual.widgets.TextArea):
    @dataclasses.dataclass
    class CursorOutBottom(textual.message.Message):
        text_area: CellTextArea

    @dataclasses.dataclass
    class CursorOutTop(textual.message.Message):
        text_area: CellTextArea

    def _on_key(self, event: textual.events.Key) -> None:
        if event.key == "escape":
            # Let it be handled upstream
            event.prevent_default()

    @tp.override
    def action_cursor_down(self, select: bool = False) -> None:
        old_location = self.cursor_location
        super().action_cursor_down(select)
        if not select and self.cursor_location == old_location:
            self.post_message(CellTextArea.CursorOutBottom(self))

    @tp.override
    def action_cursor_up(self, select: bool = False) -> None:
        old_location = self.cursor_location
        super().action_cursor_up(select)
        if not select and self.cursor_location == old_location:
            self.post_message(CellTextArea.CursorOutTop(self))


class CodeTextArea(CellTextArea):
    """Extends text area with jupyter completions."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.completer = Completer()
        self.inspect = Inspect()
        self.app: JupyterTextualApp

    def compose(self) -> textual.app.ComposeResult:
        yield self.completer
        yield self.inspect

    # TODO: type annotate
    def with_kernel(f):
        async def new_f(self, *args, **kwargs):
            if len(self.app.kernel_access_queue) > 0:
                # We don't want to wait
                self.inspect.visible = False
                self.completer.visible = False
                self.focus(scroll_visible=False)
                return

            task = self.app.queue_for_kernel(f, self, *args, **kwargs)
            try:
                await task
            except (Exception, asyncio.CancelledError):
                self.inspect.visible = False
                self.completer.visible = False
                self.focus(scroll_visible=False)

        return new_f

    def _compute_pop_offset(self, popup: PopUpMixin, virtual_cursor_position: tp.Tuple[int, int]) -> tp.Tuple[int, int]:
        start_x, start_y = self.wrapped_document.location_to_offset(virtual_cursor_position)
        estimated_height = popup.estimate_height()
        # offset_y = 1 if we put below the curser, negative otherwise
        offset_y = (
            1 if self.screen.region.height - self.cursor_screen_offset.y - estimated_height > 0 else -estimated_height
        )
        return (start_x - self.scroll_offset[0] + self.gutter_width - 1, start_y - self.scroll_offset[1] + offset_y)

    @with_kernel
    async def _complete_from_kernel(self, autoinsert: bool) -> None:
        """
        Parameters:
        -----------
        autoinsert: bool
            if True the result is automatically inserted if the completion is unique
        """
        self.inspect.visible = False
        cursor_pos = self.document.get_index_from_location(self.cursor_location)
        msg_id = self.app.kernel_client.complete(self.text, cursor_pos)
        msg = await self.app.kernel_client._recv_reply(msg_id, 5)
        self.app.log.info(f"got message {msg}")
        self.replace_start = self.document.get_location_from_index(msg["content"]["cursor_start"])
        matches = msg["content"]["matches"]
        if not matches:
            textual.log("action dismiss")
            self.completer.action_dismiss()
            textual.log(f"{self.has_focus=}")
            return
        elif len(matches) == 1 and autoinsert:
            self.replace(matches[0], self.replace_start, self.cursor_location)
            self.completer.action_dismiss()
            return
        self.completer.visible = True
        self.completer.clear_options()
        self.completer.add_options(matches)
        self.completer.highlighted = 0
        self.completer.styles.offset = self._compute_pop_offset(self.completer, self.replace_start)
        self.completer.refresh(layout=True)

    @with_kernel
    async def _show_inspect(self):
        self.completer.visible = False
        cursor_pos = self.document.get_index_from_location(self.cursor_location)
        msg_id = self.app.kernel_client.inspect(self.text, cursor_pos)
        msg = await self.app.kernel_client._recv_reply(msg_id, 5)
        self.app.log.info(f"got message {msg}")
        if "text/plain" not in msg["content"]["data"]:
            self.inspect.action_dismiss()
            return
        self.inspect.update(rich.text.Text.from_ansi(msg["content"]["data"]["text/plain"]))
        self.inspect.visible = True
        self.inspect.styles.offset = self._compute_pop_offset(self.inspect, self.cursor_location)

    def _is_start_of_line(self):
        return not self.document.get_line(self.cursor_location[0])[: self.cursor_location[1]].strip()

    async def _on_key(self, event: textual.events.Key) -> None:
        if event.key == "shift+tab":
            if not self._is_start_of_line() and self.selection.is_empty:
                event.prevent_default()
                event.stop()
                await self._show_inspect()
        elif event.key == "tab":
            if not self._is_start_of_line() and self.selection.is_empty:
                event.prevent_default()
                event.stop()
                await self._complete_from_kernel(True)
        elif self.completer.visible:
            if event.is_printable and re.match("[%0-9a-z._/\\:~-]", event.character, re.I):
                self.call_later(self._complete_from_kernel, False)
            elif event.key in self.completer.get_handle_keys():
                event.prevent_default()
            else:
                self.completer.action_dismiss()
        elif self.inspect.visible:
            if event.key in self.inspect.get_handle_keys():
                event.prevent_default()
            else:
                self.inspect.action_dismiss()

    def on_option_list_option_selected(self, message: Completer.OptionSelected) -> None:
        self.replace(message.option.prompt, self.replace_start, self.cursor_location)

    def _on_mouse_down(self, event: textual.events.MouseDown) -> None:
        if self.completer.visible:
            # The parent captures mouse which prevents the completer from getting messages
            event.prevent_default()



import asyncio
import collections
import dataclasses
import importlib
import pathlib
import sys
import time
import typing as tp

import jupyter_client

import nbformat

import rich.console

import textual.app
import textual.containers
import textual.widgets
import textual.widgets.text_area
import textual.binding
import textual._border

import tree_sitter

from netbook._cell import Cell, CodeCell, RawCell, MarkdownCell
from netbook._text_area import CellTextArea

# Monkeypatch "double" for a custom border. Not great, but will do for now.
textual._border.BORDER_CHARS["double"] = (
    ("▄", " ", " "),
    ("█", " ", " "),
    ("▀", " ", " "),
)


class JupyterTextualApp(textual.app.App):
    CSS_PATH = "netbook.css"

    COMMAND_PALETTE_BINDING = "ctrl+shift+f,ctrl+shift+p,p"

    AUTO_FOCUS = "CellTextArea"

    REPEAT_KEY_PRESS_TIME = 0.8

    CMDTRL = "super" if sys.platform == "darwin" else "ctrl"

    show_line_numbers: textual.reactive.reactive[bool] = textual.reactive.reactive(True)

    def _watch_show_line_numbers(self, show_line_numbers):
        for cell in self.cells:
            cell.source.show_line_numbers = show_line_numbers

    @dataclasses.dataclass
    class CellDeletion:
        position: int
        nbformats: list[str]

    def __init__(
        self,
        kernel_manager: jupyter_client.KernelManager,
        kernel_client: jupyter_client.AsyncKernelClient,
        nbfile: str,
        nb: nbformat.NotebookNode | None,
    ) -> None:
        super().__init__()
        self.kernel_manager = kernel_manager
        self.kernel_client = kernel_client
        self.nbfile = nbfile
        self.nb = nb or nbformat.v4.new_notebook(cells=[nbformat.v4.new_code_cell()])
        self.start_cell_id = self.focused_cell_id = 0

        self.copied_cells_nbformat: tp.List[nbformat.NotebookNode] = []
        self.cell_deletion_stack: collections.deque[self.CellDeletion] = collections.deque()

        self.kernel_access_queue: set[asyncio.Future] = set()

        self.scroll = textual.containers.Container(classes="mainscroll")
        self.cell_type_selector = textual.widgets.Select(
            [(x, x.lower()) for x in ["Code", "Markdown", "Raw"]], allow_blank=False, value=self.nb.cells[0].cell_type
        )
        self.kernel_state = textual.widgets.Static("○")

        self.repeat_key = None
        self.repeat_key_count = 0
        self.last_key_press_time = time.monotonic()

        self._load_language()
        self.call_after_refresh(lambda: self.queue_for_kernel(self._initialize_kernel))

    @property
    def cells(self) -> list[Cell]:
        return self.scroll.children

    def _load_language(self) -> None:
        self.tree_sitter_language, self.language_highlights_query = None, ""
        language = self.kernel_manager.kernel_spec.language
        ll = language.lower()
        if ll not in textual.widgets.text_area.BUILTIN_LANGUAGES:
            try:
                m = importlib.import_module(f"tree_sitter_{ll}")
                tree_sitter_language = tree_sitter.Language(m.language(), name=language)
                if not hasattr(m, "HIGHLIGHTS_QUERY"):
                    # Last ditch attempt to get the query. Works e.g. for julia
                    m._get_query("HIGHLIGHTS_QUERY", "highlights.scm")
                language_highlights_query = m.HIGHLIGHTS_QUERY
            except (ModuleNotFoundError, AttributeError, OSError):
                self.notify(
                    f"Syntax highlighting is not available for {language}. Try installing the package `tree_sitter_{ll}`"
                )
            else:
                # Only set those if everything was successful
                self.tree_sitter_language = tree_sitter_language
                self.language_highlights_query = language_highlights_query

    async def _initialize_kernel(self) -> None:
        if self.kernel_manager.kernel_spec.language == "python":
            # Enrich the output with Rich
            with open(pathlib.Path(__file__).parent.joinpath("_ipykernel_init.py")) as init_file:
                await self.kernel_client.execute_interactive(
                    init_file.read(), silent=True, output_hook=lambda msg: None
                )
        self.notify("kernel ready")

    def _kernel_task_done_callback(self, task: asyncio.Task) -> None:
        if task not in self.kernel_access_queue:
            # self.kernel_access_queue was reset. Return here so we don't mess up with the new tasks
            return
        self.kernel_access_queue.remove(task)
        if len(self.kernel_access_queue) == 0:
            self.kernel_state.update("○")
        if not task.cancelled():
            exception = task.exception()
            if exception:
                self.cancel_kernel_tasks()
            if isinstance(exception, TimeoutError):
                self.notify("kernel appears to have died, restarting")
                self.kernel_manager.restart_kernel()
                assert len(self.kernel_access_queue) == 0
                self.queue_for_kernel(self._initialize_kernel)

    def queue_for_kernel(
        self,
        task: tp.Callable[..., tp.Coroutine],
        *args,
        **kwargs,
    ) -> asyncio.Task:
        tasks_to_await = list(self.kernel_access_queue)  # Copying the state at this point

        async def _task():
            try:
                await asyncio.gather(*tasks_to_await)
            except Exception as e:
                raise asyncio.CancelledError() from e
            else:
                self.kernel_state.update("●")
                await task(*args, **kwargs)

        wrapped_task = asyncio.create_task(_task())
        wrapped_task.add_done_callback(self._kernel_task_done_callback)
        self.kernel_access_queue.add(wrapped_task)
        return wrapped_task

    def cancel_kernel_tasks(self) -> None:
        for task in self.kernel_access_queue:
            task.cancel()
        self.kernel_access_queue = set()
        self.kernel_state.update("○")

    def _get_selected_cells_range(self) -> (int, int):
        return min(self.start_cell_id, self.focused_cell_id), max(self.start_cell_id, self.focused_cell_id)

    def _adjust_cell_classes(self, fucused_cell: Cell, input_focused: bool, extend_selection: bool) -> None:
        above_focused = True
        for i, cell_ in enumerate(self.cells):
            # Since add_classes and remove_classes are expensive, we want to do it only once per cell.
            # So we try to figure out the classes in one pass.
            # The logic is a bit convoluted

            if cell_ is fucused_cell:
                self.focused_cell_id = i
                above_focused = False
                remove_classes = ["below_focused", "above_focused"]
                add_classes = ["focused"]
                if input_focused:
                    add_classes.append("input_focused")
                else:
                    remove_classes.append("input_focused")
                if not extend_selection or self.start_cell_id == self.focused_cell_id:
                    remove_classes.append("multiselect")
                else:
                    add_classes.append("multiselect")

            elif above_focused:
                remove_classes = ["focused", "input_focused", "below_focused"]
                add_classes = ["above_focused"]
                if not extend_selection or i < self.start_cell_id:
                    remove_classes.append("multiselect")
                else:
                    add_classes.append("multiselect")

            else:
                remove_classes = ["focused", "input_focused", "above_focused"]
                add_classes = ["below_focused"]
                if not extend_selection or i > self.start_cell_id:
                    remove_classes.append("multiselect")
                else:
                    add_classes.append("multiselect")

            if cell_.has_class(*add_classes):
                # add_class will not update so need to update on remove_class
                cell_.remove_class(*remove_classes)
            else:
                cell_.remove_class(*remove_classes, update=False)
                cell_.add_class(*add_classes)

        if not extend_selection:
            self.start_cell_id = self.focused_cell_id

        with self.cell_type_selector.prevent(textual.widgets.Select.Changed):
            self.cell_type_selector.value = fucused_cell.cell_type()

    async def _on_select_changed(self, message: textual.widgets.Select.Changed) -> None:
        if self.cells[self.focused_cell_id].cell_type() != message.value:
            # This function gets called at the beginning.
            # The above conditions is meant to ensure we don't call the action at the beginning.
            await self.action_change_cell_to(message.value)

    def _on_cell_text_area_cursor_out_bottom(self, message: CellTextArea.CursorOutBottom) -> None:
        if self.focused_cell_id + 1 < len(self.cells) and self.cells[self.focused_cell_id].source is message.text_area:
            next_cell = self.cells[self.focused_cell_id + 1]
            next_cell.source.move_cursor(next_cell.source.document.start)
            with next_cell.prevent(Cell.Focused):
                next_cell.action_edit_mode()
            self._focus_cell(next_cell, input_focused=True, scroll_visible=False)
            self._scroll_to_cursor(next_cell.source)

    def _on_cell_text_area_cursor_out_top(self, message: CellTextArea.CursorOutTop) -> None:
        if self.focused_cell_id > 0 and self.cells[self.focused_cell_id].source is message.text_area:
            previous_cell = self.cells[self.focused_cell_id - 1]
            previous_cell.source.move_cursor(previous_cell.source.document.end)
            with previous_cell.prevent(Cell.Focused):
                previous_cell.action_edit_mode()
            self._focus_cell(previous_cell, input_focused=True, scroll_visible=False)
            self._scroll_to_cursor(previous_cell.source)

    def _on_key(self, event: textual.events.Key) -> None:
        now = time.monotonic()
        if self.repeat_key == event.key and now - self.last_key_press_time < self.REPEAT_KEY_PRESS_TIME:
            self.repeat_key_count += 1
        else:
            self.repeat_key = event.key
            self.repeat_key_count = 1
        self.last_key_press_time = now

    def double_press(f):
        async def new_f(self, *args, **kwargs):
            if self.repeat_key_count == 2:
                self.repeat_key = None
                await f(self, *args, **kwargs)

        return new_f

    def _on_app_focus(self, event: textual.events.AppFocus) -> None:
        focused_cell = self.cells[self.focused_cell_id]
        if focused_cell.has_class("input_focused"):
            focused_cell.source.focus(False)
        else:
            focused_cell.focus(False)

    def _on_cell_focused(self, event: Cell.Focused) -> None:
        self._adjust_cell_classes(event.cell, event.input, False)

    def _on_descendant_focus(self, event: textual.events.DescendantFocus) -> None:
        if not self.scroll.has_focus_within:
            self.cells[self.focused_cell_id].remove_class("input_focused")

    def _scroll_to_cursor(self, text_area: textual.widgets.TextArea):
        # scroll self.scroll so that the cursor is visible
        region = textual.geometry.Region(text_area._cursor_offset[0], text_area._cursor_offset[1], width=3, height=1)
        widget = text_area
        while widget is not self.scroll:
            region = region.translate(widget.virtual_region.offset).translate(widget.styles.border.spacing.top_left)
            widget = widget.parent
        self.scroll.scroll_to_region(region, animate=False, immediate=True)

    def _focus_cell(
        self, cell: Cell, *, input_focused: bool = False, scroll_visible: bool = True, extend_selection: bool = False
    ) -> None:
        self._adjust_cell_classes(cell, input_focused, extend_selection)
        to_focus = cell.source if input_focused else cell
        with cell.prevent(Cell.Focused):
            self.set_focus(to_focus, scroll_visible=False)
        # scroll_visible is quite finicky. It seems like Widget.scroll_visible works best.
        if scroll_visible and not self.screen.can_view_entire(cell):
            cell.scroll_visible(immediate=True, force=True)

    def _on_text_area_selection_changed(self, event: textual.widgets.TextArea.SelectionChanged) -> None:
        self._scroll_to_cursor(event.text_area)

    @tp.override
    def exit(
        self,
        result: textual.app.ReturnType | None = None,
        return_code: int = 0,
        message: rich.console.RenderableType | None = None,
    ) -> None:
        self.kernel_client.shutdown()
        super().exit(result, return_code, message)

    def compose(self) -> textual.app.ComposeResult:
        def nf(widget: textual.widget.Widget) -> textual.widget.Widget:
            widget.can_focus = False
            return widget

        with textual.containers.Horizontal(classes="toolbar"):
            yield textual.widgets.Static(" ")
            yield nf(textual.widgets.Button("\uf0c7", action="app.save", tooltip="save"))
            yield textual.widgets.Static(" ")
            yield nf(textual.widgets.Button("\uf067", action="app.insert_cell_below", tooltip="insert cell below"))
            yield textual.widgets.Static(" ")
            yield nf(textual.widgets.Button("\uf0c4", action="app.cut_selected_cells", tooltip="cut selected cells"))
            yield nf(textual.widgets.Button("\uf0c5", action="app.copy_selected_cells", tooltip="copy selected cells"))
            yield nf(textual.widgets.Button("\uf0ea", action="app.paste_cells_below", tooltip="paste cells below"))
            yield textual.widgets.Static(" ")
            yield nf(
                textual.widgets.Button(
                    "\uf063", action="app.move_selected_cells_down", tooltip="move selected cells down"
                )
            )
            yield nf(
                textual.widgets.Button("\uf062", action="app.move_selected_cells_up", tooltip="move selected cells up")
            )
            yield textual.widgets.Static(" ")
            yield nf(
                textual.widgets.Button(
                    "\uf04b Run", action="app.run_cell_select_below", tooltip="run cell, select below"
                )
            )
            yield nf(textual.widgets.Button("\uf04d", action="app.interrupt_kernel", tooltip="interrupt the kernel"))
            yield nf(textual.widgets.Button("\uf01e", action="app.restart_kernel", tooltip="restart the kernel"))
            yield nf(
                textual.widgets.Button(
                    "\uf04e",
                    action="app.restart_and_run_all",
                    tooltip="restart the kernel, then re-run the whole notebook",
                )
            )
            yield textual.widgets.Static(" ")
            yield self.cell_type_selector
            yield textual.widgets.Static(" ")
            yield nf(textual.widgets.Button("\uf11c", action="app.command_palette"))
            yield textual.widgets.Static(" ", classes="spacer")
            with textual.containers.Horizontal(classes="title"):
                yield textual.widgets.Static(
                    f"{pathlib.Path(self.nbfile).stem} | {self.kernel_manager.kernel_spec.display_name} "
                )
                yield self.kernel_state
        with self.scroll:
            for i, cell in enumerate(self.nb.cells):
                yield Cell.from_nbformat(cell).add_class("focused" if i == 0 else "below_focused")

    BINDINGS = [
        textual.binding.Binding("f", "find_and_replace", "find and replace"),
        textual.binding.Binding(f"{CMDTRL}+shift+f,{CMDTRL}+shift+p,p", "command_palette", "open the command palette"),
        textual.binding.Binding("shift+enter", "run_cell_select_below", "run cell, select below"),
        textual.binding.Binding(f"ctrl+enter,{CMDTRL}+enter", "run_cell", "run cell"),
        textual.binding.Binding("alt+enter", "run_cell_and_insert_below", "run cell and insert below"),
        textual.binding.Binding("y", "change_cell_to('code')", "change cell to code"),
        textual.binding.Binding("m", "change_cell_to('markdown')", "change cell to markdown"),
        textual.binding.Binding("r", "change_cell_to('raw')", "change cell to raw"),
        textual.binding.Binding("k,up", "focus_cell_up", "select cell above"),
        textual.binding.Binding("j,down", "focus_cell_down", "select cell below"),
        textual.binding.Binding("K,shift+up", "extend_selected_cells_above", "extend selected cells above"),
        textual.binding.Binding("J,shift+down", "extend_selected_cells_below", "extend selected cells below"),
        textual.binding.Binding(f"{CMDTRL}+a", "select_all_cells", "select all cells"),
        textual.binding.Binding("ctrl+shift+up", "move_selected_cells_up", "move selected cells up"),
        textual.binding.Binding("ctrl+shift+down", "move_selected_cells_down", "move selected cells down"),
        textual.binding.Binding("a", "insert_cell_above", "insert cell above"),
        textual.binding.Binding("b", "insert_cell_below", "insert cell below"),
        textual.binding.Binding("x", "cut_selected_cells", "cut selected cells"),
        textual.binding.Binding("c", "copy_selected_cells", "copy selected_cells"),
        textual.binding.Binding("V", "paste_cells_above", "paste cells above"),
        textual.binding.Binding("v", "paste_cells_below", "paste cells below"),
        textual.binding.Binding("z", "undo_cell_deletion", "undo cell deletion"),
        textual.binding.Binding("d", "delete_selected_cells", "delete selected cells", key_display="d,d"),
        textual.binding.Binding(
            "M",
            "merge_selected_cells",
            "merge selected cells, or the current cell with the cell below if only one cell is selected",
        ),
        textual.binding.Binding("l", "toggle_line_numbers", "toggle line numbers"),
        textual.binding.Binding("L", "toggle_line_numbers_in_all_cells", "toggle line numbers in all cells"),
        textual.binding.Binding("o", "toggle_output", "toggle output of selected cells"),
        textual.binding.Binding("O", "toggle_output_scrolling", "toggle output scrolling of selected cells"),
        textual.binding.Binding(f"s,{CMDTRL}+s", "save", "save"),
        textual.binding.Binding("h", "toggle_help", "toggle help"),
        textual.binding.Binding("i", "try_interrupt_kernel", "interrupt the kernel", key_display="i,i"),
        textual.binding.Binding("0", "try_restart_kernel", "restart the kernel", key_display="0,0"),
        textual.binding.Binding("ctrl+shift+minus", "split_cell_at_cursor", "split cell at cursor(s)"),
    ]

    def check_action(self, action: str, parameters) -> bool | None:
        if action == "split_cell_at_cursor":
            return self.cells[self.focused_cell_id].source.has_focus_within
        return True

    def action_find_and_replace(self) -> None:
        self.notify("Not implemented yet")

    async def action_run_cell_select_below(self) -> None:
        start_id, end_id = self._get_selected_cells_range()
        for cell in self.cells[start_id : end_id + 1]:
            cell.execute()
        if end_id + 1 == len(self.cells):
            # Last cell; add a new one.
            below_cell = CodeCell(classes="below_focused")
            await self.scroll.mount(below_cell)
            self._focus_cell(below_cell, input_focused=True)
        else:
            # Not focusing the source here
            self._focus_cell(self.cells[end_id + 1], input_focused=False)

    def action_run_cell(self) -> None:
        start_id, end_id = self._get_selected_cells_range()
        for cell in self.cells[start_id : end_id + 1]:
            cell.execute()
        self._focus_cell(self.cells[end_id])

    async def action_run_cell_and_insert_below(self):
        start_id, end_id = self._get_selected_cells_range()
        for cell in self.cells[start_id : end_id + 1]:
            cell.execute()
        new_cell = CodeCell(classes="below_focused")
        await self.scroll.mount(new_cell, after=end_id)
        self._focus_cell(new_cell, input_focused=True)

    async def action_change_cell_to(self, cell_type: str) -> None:
        start_id, end_id = self._get_selected_cells_range()
        new_class = {"code": CodeCell, "raw": RawCell, "markdown": MarkdownCell}[cell_type]
        cell_to_focus = self.cells[end_id]
        self.set_focus(None)
        with self.batch_update():
            for i in range(start_id, end_id + 1):
                if not isinstance(self.cells[i], new_class):
                    new_cell = new_class(self.cells[i].source.text, classes=" ".join(self.cells[i].classes))
                    await self.scroll.mount(new_cell, after=i)
                    await self.cells[i].remove()
                    cell_to_focus = new_cell
        self._focus_cell(cell_to_focus)

    def action_focus_cell_up(self) -> None:
        if self.focused_cell_id > 0:
            self._focus_cell(self.cells[self.focused_cell_id - 1])

    def action_focus_cell_down(self) -> None:
        if self.focused_cell_id + 1 < len(self.cells):
            self._focus_cell(self.cells[self.focused_cell_id + 1])

    def action_extend_selected_cells_above(self) -> None:
        if self.focused_cell_id > 0:
            self._focus_cell(self.cells[self.focused_cell_id - 1], extend_selection=True)

    def action_extend_selected_cells_below(self) -> None:
        if self.focused_cell_id + 1 < len(self.cells):
            self._focus_cell(self.cells[self.focused_cell_id + 1], extend_selection=True)

    def action_select_all_cells(self) -> None:
        self.start_cell_id = 0
        self._focus_cell(self.cells[-1], extend_selection=True)

    def action_move_selected_cells_up(self) -> None:
        start_id, end_id = self._get_selected_cells_range()
        if start_id > 0:
            cells = list(self.cells)
            cell_to_move = cells[start_id - 1]
            cell_to_move.remove_class("above_focused", update=False).add_class("below_focused")
            cells.remove(cell_to_move)
            cells.insert(end_id, cell_to_move)
            mapping = {cell: i for i, cell in enumerate(cells)}
            self.scroll.sort_children(key=mapping.get)
            self.focused_cell_id -= 1
            self.start_cell_id -= 1
            self.screen.scroll_to_widget(self.cells[self.focused_cell_id])

    def action_move_selected_cells_down(self) -> None:
        start_id, end_id = self._get_selected_cells_range()
        if end_id + 1 < len(self.cells):
            cells = list(self.cells)
            cell_to_move = cells[end_id + 1]
            cell_to_move.remove_class("below_focused", update=False).add_class("above_focused")
            cells.remove(cell_to_move)
            cells.insert(start_id, cell_to_move)
            mapping = {cell: i for i, cell in enumerate(cells)}
            self.scroll.sort_children(key=mapping.get)
            self.focused_cell_id += 1
            self.start_cell_id += 1
            self.screen.scroll_to_widget(self.cells[self.focused_cell_id])

    async def action_insert_cell_above(self) -> None:
        start_id, end_id = self._get_selected_cells_range()
        new_cell = CodeCell(classes="above_focused")
        await self.scroll.mount(new_cell, before=start_id)
        self._focus_cell(new_cell)

    async def action_insert_cell_below(self) -> None:
        start_id, end_id = self._get_selected_cells_range()
        new_cell = CodeCell(classes="below_focused")
        await self.scroll.mount(new_cell, after=self.focused_cell_id)
        self._focus_cell(new_cell)

    def action_copy_selected_cells(self) -> None:
        start_id, end_id = self._get_selected_cells_range()
        self.copied_cells_nbformat = [cell.to_nbformat() for cell in self.cells[start_id : end_id + 1]]

    async def action_cut_selected_cells(self) -> None:
        start_id, end_id = self._get_selected_cells_range()
        self.copied_cells_nbformat = [cell.to_nbformat() for cell in self.cells[start_id : end_id + 1]]
        self.cell_deletion_stack.append(self.CellDeletion(start_id, self.copied_cells_nbformat))
        self.set_focus(None)
        await self.scroll.remove_children(self.cells[start_id : end_id + 1])
        if not self.cells:
            # Ensure there is at least one cell
            new_cell = CodeCell(classes="focused")
            await self.scroll.mount(new_cell)
        self._focus_cell(self.cells[min(start_id, len(self.cells) - 1)])

    async def action_paste_cells_below(self) -> None:
        _, end_id = self._get_selected_cells_range()
        new_cells = [Cell.from_nbformat(node).add_class("below_focused") for node in self.copied_cells_nbformat]
        await self.scroll.mount(*new_cells, after=end_id)
        if new_cells:
            self._focus_cell(new_cells[-1])

    async def action_paste_cells_above(self) -> None:
        start_id, _ = self._get_selected_cells_range()
        new_cells = [Cell.from_nbformat(node).add_class("above_focused") for node in self.copied_cells_nbformat]
        await self.scroll.mount(*new_cells, before=self.focused_cell_id)
        if new_cells:
            self._focus_cell(new_cells[0])

    async def action_undo_cell_deletion(self) -> None:
        if len(self.cell_deletion_stack) > 0:
            restore = self.cell_deletion_stack.pop()
            new_cells = [Cell.from_nbformat(nbformat) for nbformat in restore.nbformats]
            assert restore.position <= len(self.cells)
            await self.scroll.mount(*new_cells, before=restore.position)
            self._focus_cell(new_cells[-1])

    @double_press
    async def action_delete_selected_cells(self) -> None:
        start_id, end_id = self._get_selected_cells_range()
        self.cell_deletion_stack.append(
            self.CellDeletion(start_id, [cell.to_nbformat() for cell in self.cells[start_id : end_id + 1]])
        )
        self.set_focus(None)
        await self.scroll.remove_children(self.cells[start_id : end_id + 1])
        if not self.cells:
            # Ensure there is at least one cell
            new_cell = CodeCell(classes="focused")
            await self.scroll.mount(new_cell)
        self._focus_cell(self.cells[min(start_id, len(self.cells) - 1)])

    async def action_merge_selected_cells(self) -> None:
        start_id, end_id = self._get_selected_cells_range()
        if start_id == end_id:
            if end_id + 1 == len(self.cells):
                return
            # merge with the cell below
            end_id += 1
        self.cells[start_id].source.text = "\n\n".join(cell.source.text for cell in self.cells[start_id : end_id + 1])
        self.cell_deletion_stack.append(
            self.CellDeletion(start_id + 1, [cell.to_nbformat() for cell in self.cells[start_id + 1 : end_id + 1]])
        )
        await self.scroll.remove_children(self.cells[start_id + 1 : end_id + 1])
        self._focus_cell(self.cells[start_id])

    def action_save(self) -> None:
        self.nb = nbformat.v4.new_notebook(
            metadata=dict(
                kernelspec=dict(
                    name=self.kernel_manager.kernel_name,
                    language=self.kernel_manager.kernel_spec.language,
                    display_name=self.kernel_manager.kernel_spec.display_name,
                )
            ),
            cells=[cell.to_nbformat() for cell in self.cells],
        )
        nbformat.write(self.nb, self.nbfile)
        self.notify("Notebook Saved")

    def action_toggle_line_numbers(self) -> None:
        start_id, end_id = self._get_selected_cells_range()
        for cell in self.cells[start_id : end_id + 1]:
            cell.source.show_line_numbers = not cell.source.show_line_numbers

    def action_toggle_line_numbers_in_all_cells(self) -> None:
        self.show_line_numbers = not self.show_line_numbers

    def action_toggle_output(self) -> None:
        start_id, end_id = self._get_selected_cells_range()
        for cell in self.cells[start_id : end_id + 1]:
            if isinstance(cell, CodeCell):
                cell.collapsed = not cell.collapsed

    def action_toggle_output_scrolling(self) -> None:
        start_id, end_id = self._get_selected_cells_range()
        for cell in self.cells[start_id : end_id + 1]:
            if isinstance(cell, CodeCell):
                cell.scrolled = not cell.scrolled

    def action_toggle_help(self) -> None:
        if self.screen.query("HelpPanel"):
            self.action_hide_help_panel()
        else:
            self.action_show_help_panel()

    def action_interrupt_kernel(self) -> None:
        self.notify("interrupting kernel")
        self.kernel_manager.interrupt_kernel()

    @double_press
    async def action_try_interrupt_kernel(self) -> None:
        self.action_interrupt_kernel()

    def action_restart_kernel(self) -> None:
        self.notify("restarting kernel")
        # Cancell all the tasks
        self.cancel_kernel_tasks()
        self.kernel_manager.restart_kernel()
        assert len(self.kernel_access_queue) == 0
        self.queue_for_kernel(self._initialize_kernel)

    @double_press
    async def action_try_restart_kernel(self) -> None:
        self.action_restart_kernel()

    def action_restart_and_run_all(self) -> None:
        self.action_restart_kernel()
        for cell in self.cells:
            cell.execute()

    async def action_split_cell_at_cursor(self) -> None:
        focused_cell = self.cells[self.focused_cell_id]
        start = focused_cell.source.document.get_index_from_location(focused_cell.source.selection.start)
        end = focused_cell.source.document.get_index_from_location(focused_cell.source.selection.end)
        start, end = min(start, end), max(start, end)
        chunks = (focused_cell.source.text[:start], focused_cell.source.text[start:end], focused_cell.source.text[end:])

        def strip_blank_lines(s: str) -> str:
            lines = s.split("\n")
            start = 0
            while start < len(lines) and not lines[start].strip():
                start += 1
            end = len(lines) - 1
            while end >= start and not lines[end].strip():
                end -= 1
            return "\n".join(lines[start : end + 1])

        nonempty_chunks = list(filter(lambda s: len(s) > 0, map(strip_blank_lines, chunks)))

        if len(nonempty_chunks) > 1:
            new_cells = [type(focused_cell)(chunk, classes="above_focused") for chunk in nonempty_chunks[:-1]]
            await self.scroll.mount(*new_cells, before=self.focused_cell_id)
            focused_cell.source.text = nonempty_chunks[-1]
            self.focused_cell_id += len(nonempty_chunks) - 1



import textual.pilot
import netbook
import pytest_mock
import pytest


async def test_smoke(pilot: textual.pilot.Pilot):
    await pilot.exit(0)
    pilot.app.kernel_client.shutdown.assert_called_once()


async def test_save(pilot_nb: textual.pilot.Pilot, mocker: pytest_mock.MockerFixture):
    mock_write = mocker.patch("nbformat.write")
    pilot_nb.app.action_save()
    mock_write.assert_called_once()
    assert mock_write.call_args.args[1] == "./tests/test.ipynb"
    nb = mock_write.call_args.args[0]
    assert len(nb.cells) == 6
    assert tuple(cell.cell_type for cell in nb.cells) == ("code",) * 4 + ("markdown", "raw")


def test_cmdline(mocker: pytest_mock.MockerFixture):
    nb = netbook.JupyterNetbook()
    nb.initialize(["--generate-config"])
    assert not hasattr(nb, "textual_app")
    nb = netbook.JupyterNetbook()
    nb.initialize([])
    assert nb.textual_app.nbfile.startswith("Untitled")



import netbook
import asyncio
import textual


async def test_non_code_cells(pilot):
    app: netbook.JupyterTextualApp = pilot.app
    assert app.cells[0].cell_type() == "code"
    assert app.cells[0].has_class("input_focused")
    assert not app.cells[0].check_action("edit_mode", None)
    assert app.cells[0].check_action("command_mode", None)
    assert app.cells[0].check_action("notify", None)

    app.cells[0].execute()
    assert app.cells[0].execution_count == " "

    await pilot.press("escape")
    await pilot.pause()
    assert not app.cells[0].has_class("input_focused")

    await pilot.press("r")
    await pilot.pause()
    assert app.cells[0].cell_type() == "raw"
    app.cells[0].execute()

    app.cell_type_selector.value = "markdown"
    await pilot.pause()
    assert app.cells[0].cell_type() == "markdown"
    app.cells[0].execute()
    app.cells[0].source.text = "asdf"
    assert not app.cells[0].has_class("input_focused")
    await pilot.click("TextArea", times=2)
    await pilot.pause()
    assert app.cells[0].has_class("input_focused")


async def test_output_scroll_collapse(pilot_nb):
    pilot_nb: textual.pilot.Pilot
    app: netbook.JupyterTextualApp = pilot_nb.app
    assert app.cells[0].cell_type() == "code"
    assert app.cells[0].collapsed

    await pilot_nb.click(app.cells[0].expand_button)
    await pilot_nb.pause()
    assert not app.cells[0].collapsed

    await pilot_nb.click(app.cells[0].scroll_button, times=2)
    await pilot_nb.pause()
    assert app.cells[0].collapsed


async def test_run_cell(pilot, mocker):
    app: netbook.JupyterTextualApp = pilot.app
    poller = mocker.patch("zmq.asyncio.Poller")
    poller.return_value.poll = mocker.AsyncMock(return_value={app.kernel_client.iopub_channel.socket: 1})
    app.kernel_client.execute.return_value = "id1"

    # Test normal execution
    app.kernel_client.iopub_channel.get_msg = mocker.AsyncMock(
        side_effect=[
            {
                "parent_header": {"msg_id": "id1"},
                "header": {"msg_type": "stream"},
                "content": {"name": "stdout", "text": "  \nout\n  "},
            },
            {
                "parent_header": {"msg_id": "id1"},
                "header": {"msg_type": "stream"},
                "content": {"name": "stdout", "text": "outout"},
            },
            {
                "parent_header": {"msg_id": "id1"},
                "header": {"msg_type": "display_data"},
                "content": {"data": {"text/plain": "data"}, "metadata": {}},
            },
            {
                "parent_header": {"msg_id": "id1"},
                "header": {"msg_type": "execute_result"},
                "content": {"execution_count": 1, "data": {"text/plain": "data"}, "metadata": {}},
            },
            {
                "parent_header": {"msg_id": "id2"},  # Wrong id
            },
            {
                "parent_header": {"msg_id": "id1"},
                "header": {"msg_type": "status"},
                "content": {"execution_state": "idle"},
            },
        ]
    )
    app.kernel_client._recv_reply = mocker.AsyncMock(
        return_value={
            "parent_header": {"msg_id": "id1"},
            "header": {"msg_type": "execute_reply"},
            "content": {
                "status": "ok",
                "payload": [{"source": "page", "data": {"text/plain": "payload"}}],
                "execution_count": 1,
            },
        }
    )

    app.cells[0].source.text = "some code"
    app.action_run_cell()
    tasks = list(pilot.app.kernel_access_queue)
    assert len(tasks) == 1
    await asyncio.gather(*pilot.app.kernel_access_queue, return_exceptions=True)
    assert tasks[0].result() is None
    await pilot.pause()
    assert len(app.cells[0].all_outputs) == 4
    assert app.cells[0].all_outputs[0].text == "  \nout\n  outout"

    # Add a new cell
    await app.action_insert_cell_below()
    await pilot.pause()
    assert len(app.cells) == 2
    assert app.focused_cell_id == 1

    # Test error on execution
    app.cells[1].source.text = "some other code"
    app.kernel_client.execute.return_value = "id1"
    app.kernel_client.iopub_channel.get_msg = mocker.AsyncMock(
        side_effect=[
            {
                "parent_header": {"msg_id": "id1"},
                "header": {"msg_type": "error"},
                "content": {"ename": "RuntimeError", "evalue": "3", "traceback": "traceback"},
            },
            {
                "parent_header": {"msg_id": "id1"},
                "header": {"msg_type": "status"},
                "content": {"execution_state": "idle"},
            },
        ]
    )
    app.kernel_client._recv_reply = mocker.AsyncMock(
        return_value={
            "parent_header": {"msg_id": "id1"},
            "header": {"msg_type": "execute_reply"},
            "content": {"status": "error", "execution_count": 1, "payload": [], "ename": "RuntimeError"},
        }
    )

    app.action_extend_selected_cells_above()
    await pilot.pause()
    app.action_run_cell()
    tasks = list(pilot.app.kernel_access_queue)
    assert len(tasks) == 2
    await asyncio.gather(*pilot.app.kernel_access_queue, return_exceptions=True)
    assert tasks[0].cancelled() or tasks[0].exception()
    assert tasks[1].cancelled() or tasks[1].exception()
    await pilot.pause()
    assert app.cells[0].execution_count == 1
    assert app.cells[1].execution_count == " "

    # Test kernel timeout
    poller.return_value.poll = mocker.AsyncMock(side_effect=[{}, {app.kernel_client.iopub_channel.socket: 1}])
    app.kernel_manager.is_alive = mocker.Mock(side_effect=[True, False])

    app.action_select_all_cells()
    app.action_run_cell()
    app.action_select_all_cells()
    app.action_run_cell()  # Run them two times
    tasks = list(pilot.app.kernel_access_queue)
    assert len(tasks) == 4
    await asyncio.gather(*pilot.app.kernel_access_queue, return_exceptions=True)
    for task in tasks:
        assert task.cancelled() or task.exception()
    await pilot.pause()
    assert app.cells[0].execution_count == " "
    assert app.cells[1].execution_count == " "



from __future__ import annotations

import abc
import asyncio
import base64
import dataclasses
import io
import typing as tp

import nbformat

import pyvips

import rich.ansi
import rich.text

import textual.app
import textual.binding
import textual.containers
import textual.message
import textual.widgets.text_area

import textual_image.widget

import zmq

from netbook._text_area import CellTextArea, CodeTextArea

if tp.TYPE_CHECKING:
    from netbook import JupyterTextualApp


class Counter(textual.widgets.Static):
    """Displays input or output execution count."""

    execution_count: textual.reactive.reactive[int | str] = textual.reactive.reactive(" ")

    def __init__(self, format: str = "", *args, **kwargs) -> None:
        super().__init__(*args, **kwargs, markup=False)
        self.format = format

    def watch_execution_count(self, execution_count) -> None:
        self.update(self.format.format(execution_count))


class Output(textual.containers.Container):
    """Base class for code cell output."""

    @dataclasses.dataclass
    class Resized(textual.message.Message):
        """Pass on resize events to parents."""

        output: Output
        size: textual.geometry.Size

    @abc.abstractmethod
    def to_nbformat(self) -> nbformat.NotebookNode:
        raise NotImplementedError

    def on_resize(self, event: textual.events.Resize) -> None:
        self.post_message(Output.Resized(self, event.size))

    @staticmethod
    def ansi_to_rich(text: str, strip: bool = True) -> rich.text.Text:
        """Convet ansi escaped sequences to rich text.

        Work around some issues with rich.text.Text.from_ansi and optionally strip blank lines from the start and end."""

        # splitlines - used by rich.text.Text.from_ansi splits on "\r" too.
        # E.g. tqdm relies on "\r" to rerender on the same line. So we split on "\n"-only.
        lines = text.split("\n")
        start = 0
        while strip and start < len(lines) and not lines[start].strip():
            start += 1
        end = len(lines) - 1
        while strip and end >= start and not lines[end].strip():
            end -= 1
        decoder = rich.ansi.AnsiDecoder()
        return rich.text.Text("\n").join(decoder.decode_line(line) for line in lines[start : end + 1])


class Stream(Output):
    """Stream output in notebook."""

    text: textual.reactive.reactive[str] = textual.reactive.reactive("", recompose=True)

    def __init__(self, name: str, text: str = "") -> None:
        super().__init__(classes=name)
        self.stream_name = name
        self.text = text

    @tp.override
    def to_nbformat(self) -> nbformat.NotebookNode:
        return nbformat.v4.new_output(output_type="stream", name=self.stream_name, text=self.text)

    def compose(self) -> tp.Iterable[textual.widgets.Widget]:
        with textual.containers.HorizontalScroll():
            yield textual.widgets.Static(self.ansi_to_rich(self.text))


class DisplayData(Output):
    """Cell output of type display_data."""

    def __init__(self, data: dict[str, tp.Any], metadata: dict[str, tp.Any]) -> None:
        super().__init__()
        self.data, self.metadata = data, metadata

    @tp.override
    def to_nbformat(self) -> nbformat.NotebookNode:
        return nbformat.v4.new_output(output_type="display_data", data=self.data, metadata=self.metadata)

    def _find_image(self):
        for key in self.data:
            if key.startswith("image/"):
                return key
        return None

    def compose(self) -> tp.Iterable[textual.widgets.Widget]:
        with textual.containers.HorizontalScroll():
            image_key = self._find_image()
            if image_key:
                image_bytes = (
                    pyvips.Image.svgload_buffer(self.data[image_key].encode()).pngsave_buffer()
                    if image_key == "image/svg+xml"
                    else base64.b64decode(self.data[image_key])
                )
                image = textual_image.widget.Image(io.BytesIO(image_bytes))
                # We'll set the width/height explicitly since automatically setting it seems hard / impossible.
                cell_size = textual_image.widget.get_cell_size()
                image.styles.width = round(image._image_width / cell_size.width)
                image.styles.height = round(image._image_height / cell_size.height)
                yield image
            elif self.data.get("text/plain"):
                yield textual.widgets.Static(self.ansi_to_rich(self.data["text/plain"]))


class ExecuteResult(DisplayData):
    """Cell output of type execute_result."""

    def __init__(self, execution_count: int, data: dict[str, tp.Any], metadata: dict[str, tp.Any]) -> None:
        super().__init__(data, metadata)
        self.execution_count = execution_count

    @tp.override
    def to_nbformat(self) -> nbformat.NotebookNode:
        return nbformat.v4.new_output(
            output_type="execute_result", execution_count=self.execution_count, data=self.data, metadata=self.metadata
        )


class Error(Output):
    """Cell output of type error."""

    def __init__(self, ename: str, evalue: str, traceback: list[str]) -> None:
        super().__init__()
        self.ename, self.evalue, self.traceback = ename, evalue, traceback

    @tp.override
    def to_nbformat(self) -> nbformat.NotebookNode:
        return nbformat.v4.new_output(
            output_type="error", ename=self.ename, evalue=self.evalue, traceback=self.traceback
        )

    def compose(self) -> tp.Iterable[textual.widgets.Widget]:
        with textual.containers.Horizontal():
            yield textual.widgets.Static(rich.text.Text.from_ansi("\n".join(self.traceback)))


class Cell(textual.containers.Container, can_focus=True):
    """Abstract base class for a notebook cell."""

    BINDINGS = [
        textual.binding.Binding("enter", "edit_mode", "enter edit mode"),
        textual.binding.Binding("escape", "command_mode", "enter command mode"),
    ]

    @dataclasses.dataclass
    class Focused(textual.message.Message):
        cell: Cell
        input: bool = False

    @abc.abstractmethod
    def to_nbformat(self) -> nbformat.NotebookNode:
        raise NotImplementedError

    @abc.abstractmethod
    def cell_type(self) -> str:
        raise NotImplementedError

    @staticmethod
    def from_nbformat(node: nbformat.NotebookNode):
        match node.cell_type:
            case "code":
                cell = CodeCell(source=node.source)
                cell.execution_count = node.execution_count or " "
                if "metadata" in node:
                    if "collapsed" in node.metadata:
                        cell.collapsed = node.metadata.collapsed
                    if "scrolled" in node.metadata and node.metadata.scrolled != "auto":
                        cell.scrolled = node.metadata.scrolled
                for output in node.outputs:
                    match output.output_type:
                        case "stream":
                            cell.add_output(Stream(output.name, output.text))
                        case "display_data":
                            cell.add_output(DisplayData(output.data, output.metadata))
                        case "execute_result":
                            cell.add_output(ExecuteResult(output.execution_count, output.data, output.metadata))
                        case "error":
                            cell.add_output(Error(output.ename, output.evalue, output.traceback))
                        case _:
                            raise RuntimeError(f"Unknown output type {output.output_type}")
            case "markdown":
                cell = MarkdownCell(source=node.source)
            case "raw":
                cell = RawCell(source=node.source)
            case _:
                raise RuntimeError(f"Unknown cell type {node.cell_type}")
        return cell

    def __init__(self, source: textual.widgets.TextArea, *, classes: str | None = None) -> None:
        """Constructor.

        Parameters
        ----------
        source: TextArea
            Editor widget for this cell.
        """
        self.app: JupyterTextualApp
        super().__init__(classes=classes)
        self.source = source

    def on_descendant_focus(self, event: textual.events.DescendantFocus) -> None:
        self.post_message(Cell.Focused(self, input=self.source.has_focus_within))

    def on_focus(self, message: textual.events.Focus):
        self.post_message(Cell.Focused(self))

    def check_action(self, action: str, parameters) -> bool | None:
        if action == "edit_mode":
            return not self.source.has_focus_within
        if action == "command_mode":
            return self.source.has_focus_within
        return True

    def action_edit_mode(self) -> None:
        self.source.focus(scroll_visible=False)

    def action_command_mode(self) -> None:
        self.focus(scroll_visible=False)

    @abc.abstractmethod
    def execute(self) -> None:
        """Execute this cell."""
        raise NotImplementedError


class MarkdownCell(Cell):
    edit_mode: textual.reactive.reactive[bool] = textual.reactive.reactive(True, layout=True)

    def watch_edit_mode(self, new_value: bool) -> None:
        if new_value:
            self.source.remove_class("hidden")
            self.markdown.add_class("hidden")
        else:
            self.markdown.update(self.source.text)
            self.source.add_class("hidden")
            self.markdown.remove_class("hidden")

    def __init__(self, source: str = "", *, classes: str | None = None) -> None:
        super().__init__(
            CellTextArea.code_editor(source, language="markdown", soft_wrap=True, highlight_cursor_line=False),
            classes=classes,
        )
        self.markdown = textual.widgets.Markdown()

    def on_text_area_changed(self, event: textual.widgets.TextArea.Changed):
        if not self.edit_mode:
            self.markdown.update(event.text_area.text)

    @tp.override
    def to_nbformat(self) -> nbformat.NotebookNode:
        return nbformat.v4.new_markdown_cell(source=self.source.text)

    @tp.override
    def cell_type(self) -> str:
        return "markdown"

    def compose(self) -> tp.Iterable[textual.widgets.Widget]:
        with textual.containers.Horizontal():
            yield Counter()
            yield self.source
            yield self.markdown

    def on_click(self, message: textual.events.Click) -> None:
        if message.chain > 1:
            self.action_edit_mode()

    @tp.override
    def execute(self) -> None:
        self.edit_mode = False

    @tp.override
    def action_edit_mode(self) -> None:
        self.edit_mode = True
        super().action_edit_mode()


class DoubleClickButton(textual.widgets.Button):
    """A button that can be double clicked."""

    @dataclasses.dataclass
    class DoublePressed(textual.events.Message):
        button: DoubleClickButton

    def __init__(self, label: str, *, tooltip: str) -> None:
        super().__init__(label, tooltip=tooltip)

    async def on_click(self, event: textual.events.Click) -> None:
        event.stop()
        event.prevent_default()
        if not self.has_class("-active") and event.chain == 1:
            self.press()
        elif event.chain == 2:
            # We don't check for "-active" here since on_click is called twice for double click.
            self.double_press()

    def double_press(self) -> tp.Self:
        if self.display and not self.disabled:
            self._start_active_affect()
            self.post_message(DoubleClickButton.DoublePressed(self))
        return self


class CodeCell(Cell):
    scrolled: textual.reactive.reactive[bool | None] = textual.reactive.reactive(None)  # None for auto
    collapsed: textual.reactive.reactive[bool] = textual.reactive.reactive(False)
    execution_count: textual.reactive.reactive[int | str] = textual.reactive.reactive(" ")

    def watch_scrolled(self, scrolled: bool | None) -> None:
        if scrolled or (scrolled is None and sum(self.output_heights.values()) > 105):
            self.output_area.add_class("scrolled")
            self.scroll_button.tooltip = "click to unscroll output; double click to hide"
        else:
            self.output_area.remove_class("scrolled")
            self.scroll_button.tooltip = "click to scroll output; double click to hide"

    def watch_collapsed(self, collapsed: bool) -> None:
        if collapsed:
            self.output_area.remove_class("noncollapsed")
            self.output_area.add_class("collapsed")
        else:
            self.output_area.remove_class("collapsed")
            self.output_area.add_class("noncollapsed")

    def __init__(self, source: str = "", *, classes: str | None = None) -> None:
        language = self.app.kernel_manager.kernel_spec.language.lower()
        initial_language = language if language in textual.widgets.text_area.BUILTIN_LANGUAGES else None
        text_area = CodeTextArea.code_editor(source, language=initial_language, highlight_cursor_line=False)
        if language not in textual.widgets.text_area.BUILTIN_LANGUAGES and self.app.language_highlights_query:
            text_area.register_language(language, self.app.tree_sitter_language, self.app.language_highlights_query)
            text_area.language = language

        super().__init__(text_area, classes=classes)
        self.output_area = textual.containers.Container(classes="codeoutput noncollapsed")
        self.all_outputs: list[Output] = []
        self.all_outputs_container = textual.containers.Vertical()
        self.scroll_button = DoubleClickButton("", tooltip="click to scroll output; double click to hide")
        self.expand_button = textual.widgets.Button("...", tooltip="click to expand output")
        self.scroll_button.styles.line_pad = 0  # For some reason can't set this in css
        self.output_heights = {}
        self.n_active_executions = 0

    def on_double_click_button_double_pressed(self, event: DoubleClickButton.DoublePressed):
        assert not self.collapsed
        self.collapsed = True
        self.focus(False)

    def on_button_pressed(self, event: textual.widgets.Button.Pressed):
        match event.button:
            case self.scroll_button:
                self.scrolled = not self.output_area.has_class("scrolled")
            case self.expand_button:
                assert self.collapsed
                self.collapsed = False
                self.scrolled = None
                self.focus(False)
            case _:
                assert False

    @tp.override
    def to_nbformat(self) -> nbformat.NotebookNode:
        return nbformat.v4.new_code_cell(
            source=self.source.text,
            execution_count=self.execution_count if isinstance(self.execution_count, int) else None,
            outputs=[output.to_nbformat() for output in self.all_outputs],
            metadata={"collapsed": self.collapsed, "scrolled": "auto" if self.scrolled is None else self.scrolled},
        )

    @tp.override
    def cell_type(self) -> str:
        return "code"

    def on_output_resized(self, message: Output.Resized) -> None:
        self.output_heights[message.output] = message.size.height

        # update button label
        label = ""
        for output in self.all_outputs:
            if isinstance(output, ExecuteResult):
                label += f"Out\\[{output.execution_count}]:"
                break
            else:
                label += "\n" * self.output_heights.get(output, 0)
        if label.strip():
            self.scroll_button.label = label

        # update scrolled if necessary
        self.watch_scrolled(self.scrolled)

    def add_output(self, output: Output) -> textual.widget.AwaitMount:
        if self.n_active_executions <= 1:
            self.all_outputs.append(output)
            if self.is_mounted:
                return self.all_outputs_container.mount(output)
            else:
                return textual.widget.AwaitMount(self, [])

    def compose(self) -> tp.Iterable[textual.widgets.Widget]:
        with textual.containers.Horizontal():
            yield Counter("In [{}]:", classes="input").data_bind(CodeCell.execution_count)
            yield self.source
        with self.output_area:
            yield self.expand_button
            with textual.containers.Horizontal():
                yield self.scroll_button
                with self.all_outputs_container:
                    for output in self.all_outputs:
                        yield output

    async def _parse_message(self, message: dict[str, tp.Any]) -> None:
        self.app.log.info(f"got message {message}")
        header = message["header"]
        content = message["content"]
        await_mount = None
        match header["msg_type"]:
            case "execute_result":
                await_mount = self.add_output(
                    ExecuteResult(content["execution_count"], content["data"], content["metadata"])
                )
            case "display_data":
                await_mount = self.add_output(DisplayData(content["data"], content["metadata"]))
            case "execute_reply":
                if self.n_active_executions == 1:
                    # if the cell has been rerun, don't set the execution_count
                    self.execution_count = content["execution_count"]
                for payload in content["payload"]:
                    match payload["source"]:
                        case "page":
                            await_mount = self.add_output(DisplayData(payload["data"], {}))
                if content["status"] != "ok":
                    raise RuntimeError(content["ename"])
            case "stream":
                if (
                    self.all_outputs
                    and isinstance(self.all_outputs[-1], Stream)
                    and self.all_outputs[-1].stream_name == content["name"]
                ):
                    self.all_outputs[-1].text = self.all_outputs[-1].text + content["text"]
                else:
                    await_mount = self.add_output(Stream(name=content["name"], text=content["text"]))
            case "error":
                await_mount = self.add_output(Error(content["ename"], content["evalue"], content["traceback"]))
        if await_mount:
            await await_mount

            def scroll_callback(widget):
                if not self.screen.can_view_entire(widget):
                    self.screen.scroll_to_widget(widget, center=True, force=True)

            self.all_outputs[-1].call_after_refresh(scroll_callback, self.all_outputs[-1])

    @tp.override
    def execute(self) -> None:
        self.output_heights = {}
        # TODO: do we need to await this?
        self.all_outputs_container.remove_children()
        self.all_outputs = []
        self.collapsed = False
        self.scroll_button.label = ""
        if not self.source.text:
            self.execution_count = " "
            return
        self.n_active_executions += 1
        self.execution_count = "*"

        def task_done(task: asyncio.Task):
            self.n_active_executions -= 1
            if self.execution_count == "*" and (task.cancelled() or task.exception()):
                self.execution_count = " "

        self.app.queue_for_kernel(self._execute).add_done_callback(task_done)

    async def _execute(self) -> None:
        kernel_client = self.app.kernel_client

        msg_id = kernel_client.execute(self.source.text)

        poller = zmq.asyncio.Poller()
        iopub_socket = kernel_client.iopub_channel.socket
        poller.register(iopub_socket, zmq.POLLIN)

        # wait for output and redisplay it
        while True:
            if not self.app.kernel_manager.is_alive():
                raise TimeoutError

            events = dict(await poller.poll(timeout=1))
            if iopub_socket not in events:
                continue

            msg = await kernel_client.iopub_channel.get_msg(timeout=1)

            if msg["parent_header"].get("msg_id") != msg_id:
                # not from my request
                continue
            await self._parse_message(msg)

            # stop on idle
            if msg["header"]["msg_type"] == "status" and msg["content"]["execution_state"] == "idle":
                break

        # output is done, get the reply
        res = await kernel_client._recv_reply(msg_id, timeout=1)
        await self._parse_message(res)


class RawCell(Cell):
    def __init__(self, source: str = "", *, classes: str | None = None) -> None:
        super().__init__(CodeTextArea.code_editor(source, soft_wrap=True), classes=classes)

    @tp.override
    def to_nbformat(self) -> nbformat.NotebookNode:
        return nbformat.v4.new_raw_cell(source=self.source.text)

    @tp.override
    def cell_type(self) -> str:
        return "raw"

    def compose(self) -> tp.Iterable[textual.widgets.Widget]:
        with textual.containers.Horizontal():
            yield Counter()
            yield self.source

    @tp.override
    def execute(self) -> None:
        pass



import netbook
import textual.pilot
import pytest_mock
import nbformat


async def test_actions(pilot: textual.pilot.Pilot, mocker: pytest_mock.MockerFixture):
    app: netbook.JupyterTextualApp = pilot.app

    # Test split cell
    assert len(app.cells) == 1
    assert app.cells[0].has_class("input_focused")
    await pilot.press("a", "enter", "enter", "b", "up", "ctrl+shift+minus")
    await pilot.pause()
    assert len(app.cells) == 2
    assert app.cells[0].source.text == "a"
    assert app.cells[1].source.text == "b"
    assert app.cells[1].has_class("input_focused")

    # Test run cell
    await pilot.press("up")
    await pilot.pause()
    execute_mock = mocker.Mock()
    mocker.patch("netbook._cell.CodeCell.execute", execute_mock)
    await pilot.press("shift+enter")
    await pilot.pause()
    execute_mock.assert_called_once()
    assert len(app.cells) == 2
    assert app.cells[1].has_class("focused")
    assert not app.cells[1].has_class("input_focused")

    # Test up/down
    await pilot.press("up")
    await pilot.pause()
    assert app.cells[0].has_class("focused")
    await pilot.press("down")
    await pilot.pause()
    assert app.cells[1].has_class("focused")
    await pilot.press("shift+up")
    await pilot.pause()
    assert app.cells[0].has_class("focused")
    assert app.cells[0].has_class("multiselect")
    await pilot.press("shift+down")
    assert app.cells[1].has_class("focused")
    assert not app.cells[1].has_class("multiselect")

    # Test move up/down
    await pilot.press("ctrl+shift+up")
    await pilot.pause()
    assert app.cells[0].has_class("focused")
    assert app.cells[0].source.text == "b"
    assert app.cells[1].source.text == "a"
    await pilot.press("ctrl+shift+down")
    await pilot.pause()
    assert app.cells[0].source.text == "a"
    assert app.cells[1].source.text == "b"

    # Insert cells
    assert app.cells[1].has_class("focused")
    assert not app.cells[1].has_class("input_focused")
    await pilot.press("a")
    await pilot.pause()
    assert len(app.cells) == 3
    assert app.cells[2].source.text == "b"
    assert app.cells[1].has_class("focused")
    assert not app.cells[1].has_class("input_focused")
    await pilot.press("b")
    await pilot.pause()
    assert len(app.cells) == 4
    assert app.cells[2].has_class("focused")

    # Merge cells
    await pilot.press("M", "M")
    await pilot.pause()
    assert len(app.cells) == 3
    assert app.cells[2].source.text == "\n\nb"
    assert app.cells[2].has_class("focused")


async def test_copy_paste(pilot: textual.pilot.Pilot):
    app: netbook.JupyterTextualApp = pilot.app
    await pilot.press("a", "s", "d", "f", "escape")
    await pilot.pause()
    assert len(app.cells) == 1
    assert app.cells[0].source.text == "asdf"

    # Test cut
    await pilot.press("x")
    await pilot.pause()
    # should insert a new empty cell
    assert len(app.cells) == 1
    assert app.cells[0].source.text == ""

    # Test paste
    await pilot.press("v")
    await pilot.pause()
    assert len(app.cells) == 2
    assert app.cells[1].source.text == "asdf"
    assert app.cells[1].has_class("focused")

    # Test copy and paste above
    await pilot.press("up", "c", "V")
    await pilot.pause()
    assert len(app.cells) == 3
    assert app.cells[0].has_class("focused")
    assert app.cells[0].source.text == ""

    # Test delete
    await pilot.press(f"{app.CMDTRL}+a", "d", "d")
    await pilot.pause()
    # should insert a new empty cell
    assert len(app.cells) == 1
    assert app.cells[0].source.text == ""

    # Test undo deletion
    await pilot.press("z")
    await pilot.pause()
    assert len(app.cells) == 4
    assert app.cells[0].source.text == ""
    assert app.cells[1].source.text == ""
    assert app.cells[2].source.text == "asdf"
    assert app.cells[3].source.text == ""

    # One more undo
    await pilot.press("z")
    await pilot.pause()
    assert len(app.cells) == 5
    assert app.cells[0].source.text == "asdf"


async def test_toggle_actions(pilot: textual.pilot.Pilot):
    app: netbook.JupyterTextualApp = pilot.app

    # Go into command mode
    await pilot.press("escape")
    await pilot.pause()
    assert app.cells[0].source.show_line_numbers

    # Toggle line numbers
    await pilot.press("l")
    await pilot.pause()
    assert not app.cells[0].source.show_line_numbers

    # Add a new cell
    await pilot.press("b")
    await pilot.pause()
    assert app.cells[1].source.show_line_numbers

    # Toggle all line numbers
    await pilot.press("L")
    await pilot.pause()
    assert not app.cells[0].source.show_line_numbers
    assert not app.cells[1].source.show_line_numbers
    await pilot.press("L")
    await pilot.pause()
    assert app.cells[0].source.show_line_numbers
    assert app.cells[1].source.show_line_numbers

    assert app.cells[1].has_class("focused")
    assert not app.cells[1].collapsed
    assert not app.cells[1].scrolled

    # Toggle output
    await pilot.press("o")
    await pilot.pause()
    assert app.cells[1].collapsed
    assert not app.cells[1].scrolled

    # Toggle output scrolling
    await pilot.press("O")
    await pilot.pause()
    assert app.cells[1].scrolled

    # Toggle help
    assert not app.screen.query("HelpPanel")
    await pilot.press("h")
    await pilot.pause()
    assert app.screen.query("HelpPanel")
    await pilot.press("h")
    await pilot.pause()
    assert not app.screen.query("HelpPanel")


async def test_kernel_actions(pilot: textual.pilot.Pilot, mocker: pytest_mock.MockerFixture):
    app: netbook.JupyterTextualApp = pilot.app
    assert app.cells[0].has_class("input_focused")
    await pilot.press("a", "s", "df", "escape")
    await pilot.pause()

    # Interrupt kernel
    await pilot.press("i", "i")
    await pilot.pause()
    app.kernel_manager.interrupt_kernel.assert_called_once()

    # Restart kernel
    await pilot.press("0", "0")
    await pilot.pause()
    app.kernel_manager.restart_kernel.assert_called_once()

    # Restart and run all cells
    mock_execute = mocker.Mock()
    mocker.patch("netbook._cell.CodeCell.execute", mock_execute)
    app.action_restart_and_run_all()
    assert app.kernel_manager.restart_kernel.call_count == 2
    mock_execute.assert_called_once()


async def test_load_language(mocker: pytest_mock.MockerFixture):
    # Test loading lanugage with available tree sitter package
    nb = nbformat.v4.new_notebook(cells=[nbformat.v4.new_code_cell()])
    km = mocker.Mock()
    km.kernel_spec.language = "julia"
    kc = mocker.Mock()
    kc.execute_interactive = mocker.AsyncMock()
    app = netbook.JupyterTextualApp(km, kc, "", nb)
    assert app.tree_sitter_language is not None
    assert app.language_highlights_query != ""
    async with app.run_test() as pilot:
        pilot.pause()
        assert app.cells[0].source.language == "julia"

    # Test when we can't load the language
    km.kernel_spec.language = "R"
    mocker.patch("netbook._textual_app.JupyterTextualApp.notify")
    app = netbook.JupyterTextualApp(km, kc, "", nb)
    app.notify.assert_called_once_with(
        "Syntax highlighting is not available for R. Try installing the package `tree_sitter_r`"
    )
